package main

import (
	"archive/tar"
	"flag"
	"fmt"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"time"

	"github.com/cenkalti/backoff"
	"github.com/ulikunitz/xz"
)

func fetchBundle() {
	fetchBundleCmd := flag.NewFlagSet("fetchBundle", flag.ExitOnError)
	url := fetchBundleCmd.String("url", "", "URL to compiler bundle")
	targetDir := fetchBundleCmd.String("targetDir", "", "Bundle target directory")

	fetchBundleCmd.Parse(os.Args[2:])
	if fetchBundleCmd.Parsed() {
		if *url == "" {
			fmt.Printf("URL is required\n")
			os.Exit(2)
		}
		if *targetDir == "" {
			fmt.Printf("Target Dir is required\n")
			os.Exit(2)
		}
		downloadAndUnpack(*url, *targetDir)
	} else {
		os.Exit(4)
	}
}

func newUserFriendlyBackoff() backoff.BackOff {
	b := backoff.NewExponentialBackOff()
	b.MaxInterval = 5 * time.Second
	b.MaxElapsedTime = 5 * time.Second

	return b
}

func downloadAndUnpack(url, targetDir string) {
	var res *http.Response
	getBootstrapStream := func() error {
		var err error
		res, err = http.Get(url)
		if err != nil {
			return err
		}
		if res.StatusCode >= 400 && res.StatusCode < 500 {
			return backoff.Permanent(fmt.Errorf("Permanent error while fetching bootstrap package: %v", err))
		}
		if res.StatusCode != 200 {
			return fmt.Errorf("Got HTTP %v", res.StatusCode)
		}
		return nil
	}
	err := backoff.Retry(getBootstrapStream, newUserFriendlyBackoff())
	if err != nil {
		fmt.Printf("Error while fetching bootstrap package: %v", err)
		os.Exit(3)
	}
	defer res.Body.Close()
	r, err := xz.NewReader(res.Body)
	if err != nil {
		fmt.Printf("Invalid compression header: %v", err)
		os.Exit(3)
	}
	if _, err := os.Stat(targetDir); !os.IsNotExist(err) {
		err = os.RemoveAll(targetDir)
		if err != nil {
			fmt.Printf("Couldn't delete targetDir")
			os.Exit(3)
		}
	}

	tr := tar.NewReader(r)
tarUnpackLoop:
	for {
		header, err := tr.Next()

		switch {
		case err == io.EOF:
			break tarUnpackLoop

		case err != nil:
			fmt.Printf("Failed to extract: %v", err)
			os.Exit(3)

		case header == nil:
			continue
		}

		target := filepath.Join(targetDir, header.Name)

		switch header.Typeflag {
		case tar.TypeDir:
			if _, err := os.Stat(target); err != nil {
				if err := os.MkdirAll(target, 0755); err != nil {
					fmt.Printf("Failed to create directory to extract: %v", err)
					os.Exit(3)
				}
			}
		case tar.TypeReg:
			f, err := os.OpenFile(target, os.O_CREATE|os.O_RDWR, os.FileMode(header.Mode))
			if err != nil {
				fmt.Printf("Failed to create file to extract: %v", err)
				os.Exit(3)
			}
			if _, err := io.Copy(f, tr); err != nil {
				fmt.Printf("Failed to stream extraction: %v", err)
				os.Exit(3)
			}
			f.Close()
		}
	}
	fmt.Printf("Fetched Bundle without errors")
}
