package main

import (
	"sort"
	"strings"

	"path/filepath"

	"github.com/fatih/color"

	"bytes"
	"encoding/json"
	"flag"
	"fmt"

	// "io/ioutil"
	"net"
	"os"
)

var (
	colorInfo           = color.New(color.FgYellow)
	colorInfoIdent      = color.New(color.FgHiYellow)
	colorError          = color.New(color.FgRed)
	colorErrorIdent     = color.New(color.FgHiRed)
	colorCorrect        = color.New(color.FgGreen)
	colorCorrectIdent   = color.New(color.FgHiGreen)
	colorProgramOutput  = color.New(color.FgBlue)
	colorExpectedOutput = color.New(color.FgCyan)
)

func sendResult(conn net.Conn, binaryPath, sampleName string, res sampleExecutionResult) {
	type transportSampleExecutionResult struct {
		Binarypath string
		Samplename string
		Result     sampleExecutionResult
	}

	msg, err := json.Marshal(transportSampleExecutionResult{
		Binarypath: binaryPath,
		Samplename: sampleName,
		Result:     res,
	})
	if err != nil {
		panic(err)
	}
	fmt.Fprintf(conn, string(msg))
	fmt.Fprintf(conn, "\n")
}

func printRunManualInput(binaryPath string) {
	colorInfo.Printf("Running program ' ")
	colorInfoIdent.Printf("%v", binaryPath)
	colorInfo.Printf("' with your input :\n")
}
func printRunSampleInfo(binaryPath, sampleName string) {
	colorInfo.Printf("Running sample '")
	colorInfoIdent.Printf("%v", sampleName)
	colorInfo.Printf("' on program '")
	colorInfoIdent.Printf("%v", binaryPath)
	colorInfo.Printf("':\n")
}
func printSampleVerdict(binaryPath, sampleName, programOutput, expectedOutput, processError string) {
	if programOutput != expectedOutput {
		// Todo: Highlight diff
		colorError.Printf("Sample '")
		colorErrorIdent.Printf("%v", sampleName)
		colorError.Printf("': Output differs\n")
		colorInfo.Printf("Expected output:\n")
		colorProgramOutput.Printf("%v", expectedOutput)
		colorInfo.Printf("\nYour program's output:\n")
		colorExpectedOutput.Printf("%v\n", programOutput)
	} else {
		colorCorrect.Printf("Sample '")
		colorCorrectIdent.Printf("%v", sampleName)
		colorCorrect.Printf("': Correct\n")
	}
	fmt.Printf("%v\n", processError)
}

type anything interface {
}

func mapToSortedKeys(m map[string]Sample) []string {
	keys := []string{}
	for k := range m {
		keys = append(keys, k)
	}
	sort.Strings(keys)
	return keys
}

func normalizeString(in string) string {
	in = strings.ReplaceAll(in, "\r\n", "\n")
	return strings.TrimSpace(in)
}

func runOneSample() {
	runOneSampleCmd := flag.NewFlagSet("runOneSample", flag.ExitOnError)
	binaryPath := runOneSampleCmd.String("binaryPath", "", "path of binary which is executed")
	soitaskPath := runOneSampleCmd.String("soitaskPath", "", "path of soitask file, that contains the sample")
	sampleName := runOneSampleCmd.String("sampleName", "", "Full path of the sample which is executed. E.g. sample.01")
	port := runOneSampleCmd.String("port", "", "Port to which the runner reports the evaluation reports")

	runOneSampleCmd.Parse(os.Args[2:])
	if runOneSampleCmd.Parsed() {
		if *binaryPath == "" {
			fmt.Printf("binaryPath is required\n")
			os.Exit(2)
		}
		if *soitaskPath == "" {
			fmt.Printf("soitaskPath is required\n")
			os.Exit(2)
		}
		if *sampleName == "" {
			fmt.Printf("sampleName is required\n")
			os.Exit(2)
		}
		if *port == "" {
			fmt.Printf("port is required\n")
			os.Exit(2)
		}
	} else {
		os.Exit(4)
	}
	conn, err := net.Dial("tcp", "127.0.0.1:"+*port)
	if err != nil {
		panic(err)
	}
	dir, err := os.Getwd()
	if err != nil {
		fmt.Printf("%v", err)
		os.Exit(5)
	}
	fullBinaryPath := filepath.Join(dir, *binaryPath)
	fullSoitaskPath := filepath.Join(dir, *soitaskPath)

	sample, err := parseSample(fullSoitaskPath, *sampleName)
	if err != nil {
		panic(err)
	}
	input := normalizeString(sample.Input)

	printRunSampleInfo(*binaryPath, *sampleName)
	executionResult := runSample(fullBinaryPath, bytes.NewBufferString(input), os.Stdout, os.Stderr)
	expectedOutput := normalizeString(sample.Output)
	gotOutput := normalizeString(executionResult.Stdout)
	gotStderr := normalizeString(executionResult.Stderr)
	printSampleVerdict(fullBinaryPath, *sampleName, gotOutput, expectedOutput, gotStderr)

	sendResult(conn, *binaryPath, *sampleName, executionResult)

}

func runAllSamples() {
	runAllSampleCmd := flag.NewFlagSet("runAllSamples", flag.ExitOnError)
	binaryPath := runAllSampleCmd.String("binaryPath", "", "path of binary which is executed")
	soitaskPath := runAllSampleCmd.String("soitaskPath", "", "path of soitask file, that contains the sample")
	port := runAllSampleCmd.String("port", "", "Port to which the runner reports the evaluation reports")

	runAllSampleCmd.Parse(os.Args[2:])
	if runAllSampleCmd.Parsed() {
		if *binaryPath == "" {
			fmt.Printf("binaryPath is required\n")
			os.Exit(2)
		}
		if *soitaskPath == "" {
			fmt.Printf("soitaskPath is required\n")
			os.Exit(2)
		}
		if *port == "" {
			fmt.Printf("port is required\n")
			os.Exit(2)
		}
	} else {
		os.Exit(4)
	}
	conn, err := net.Dial("tcp", "127.0.0.1:"+*port)
	if err != nil {
		panic(err)
	}

	dir, err := os.Getwd()
	if err != nil {
		fmt.Printf("%v", err)
		os.Exit(5)
	}
	fullBinaryPath := filepath.Join(dir, *binaryPath)
	fullSoitaskPath := filepath.Join(dir, *soitaskPath)

	samples, err := parseAllSamples(fullSoitaskPath)
	if err != nil {
		panic(err)
	}
	for _, sampleName := range mapToSortedKeys(samples) {
		sample := samples[sampleName]

		input := normalizeString(sample.Input)
		printRunSampleInfo(*binaryPath, sampleName)
		// executionResult := runSample(*binaryPath, bytes.NewBufferString(sample.Input), ioutil.Discard, ioutil.Discard)

		executionResult := runSample(fullBinaryPath, bytes.NewBufferString(input), os.Stdout, os.Stderr)
		expectedOutput := normalizeString(sample.Output)
		gotOutput := normalizeString(executionResult.Stdout)
		gotStderr := normalizeString(executionResult.Stderr)
		printSampleVerdict(fullBinaryPath, sampleName, gotOutput, expectedOutput, gotStderr)

		sendResult(conn, *binaryPath, sampleName, executionResult)
	}

}

func runWithManualInput() {
	runWithManualInputCmd := flag.NewFlagSet("runWithManualInput", flag.ExitOnError)
	binaryPath := runWithManualInputCmd.String("binaryPath", "", "path of binary which is executed")
	port := runWithManualInputCmd.String("port", "", "Port to which the runner reports the evaluation reports")

	runWithManualInputCmd.Parse(os.Args[2:])
	if runWithManualInputCmd.Parsed() {
		if *binaryPath == "" {
			fmt.Printf("binaryPath is required\n")
			os.Exit(2)
		}
		if *port == "" {
			fmt.Printf("port is required\n")
			os.Exit(2)
		}
	} else {
		os.Exit(4)
	}
	conn, err := net.Dial("tcp", "127.0.0.1:"+*port)
	if err != nil {
		panic(err)
	}

	dir, err := os.Getwd()
	if err != nil {
		fmt.Printf("%v", err)
		os.Exit(5)
	}
	fullBinaryPath := filepath.Join(dir, *binaryPath)

	printRunManualInput(*binaryPath)
	executionResult := runSample(fullBinaryPath, os.Stdin, os.Stdout, os.Stderr)

	sendResult(conn, *binaryPath, "manualInput", executionResult)
}
