"use strict";
import * as vscode from "vscode";
import { TaskTreeViewProvider } from "./taskTreeView";
import { SamplesTreeViewProvider } from "./samplesTreeView";

import { initializeWorkspace } from "./soiworkspace";
import { createTask, createSoitaskFile, createCppFile } from "./soitask";
import {
	insertCppTemplate,
	subscribeInsertCppTemplateLens
} from "./lenses/emptyCppLens";
import {
	subscribeInsertSoitaskTemplateLens,
	insertSoitaskTemplate
} from "./lenses/emptySoitaskLens";
import { Runner } from "./runner";
import { Store } from "./store";

import { HelperBinaryResultReceiver } from "./helperBinaryResultReceiver";
import { gotoSample } from "./sample";
import { downloadCompilerBundle } from "./helperBinary";

const port = 16314;

export async function activate(context: vscode.ExtensionContext) {
	let store = new Store(context);
	let runner = new Runner(store, port);

	let createTaskCommand = vscode.commands.registerCommand(
		"soicode.createTask",
		createTask.bind(undefined, store)
	);
	context.subscriptions.push(createTaskCommand);

	let initializeWorkspaceCommand = vscode.commands.registerCommand(
		"soicode.initializeWorkspace",
		initializeWorkspace.bind(undefined, store)
	);
	context.subscriptions.push(initializeWorkspaceCommand);

	let openSampleCommand = vscode.commands.registerCommand(
		"soicode.openSample",
		gotoSample.bind(undefined, store)
	);
	context.subscriptions.push(openSampleCommand);

	let runSolutionOnSampleCommand = vscode.commands.registerCommand(
		"soicode.runSolutionOnSample",
		runner.runOneSample.bind(runner)
	);
	context.subscriptions.push(runSolutionOnSampleCommand);

	let runSolutionOnAllSamplesCommand = vscode.commands.registerCommand(
		"soicode.runSolutionOnAllSamples",
		runner.runAllSamples.bind(runner)
	);
	context.subscriptions.push(runSolutionOnAllSamplesCommand);

	let runSolutionWithManualInputCommand = vscode.commands.registerCommand(
		"soicode.runSolutionWithManualInput",
		runner.runWithManualInput.bind(runner)
	);
	context.subscriptions.push(runSolutionWithManualInputCommand);

	let debugSolutionOnSampleCommand = vscode.commands.registerCommand(
		"soicode.debugSolutionOnSample",
		debugSolutionOnSample
	);
	context.subscriptions.push(debugSolutionOnSampleCommand); // Todo

	let debugSolutionWithManualInputCommand = vscode.commands.registerCommand(
		"soicode.debugSolutionWithManualInput",
		runner.askAndDebugManually.bind(runner)
	);
	context.subscriptions.push(debugSolutionWithManualInputCommand); // Todo

	let compileCurrentSolutionCommand = vscode.commands.registerCommand(
		"soicode.compileCurrentSolution",
		runner.askAndCompile.bind(runner)
	);
	context.subscriptions.push(compileCurrentSolutionCommand);

	let addSoitaskTemplateCommand = vscode.commands.registerCommand(
		"soicode.addSoitaskTemplate",
		insertSoitaskTemplate.bind(undefined)
	);
	context.subscriptions.push(addSoitaskTemplateCommand);

	let addCppTemplateCommand = vscode.commands.registerCommand(
		"soicode.addCppTemplate",
		insertCppTemplate.bind(undefined)
	);
	context.subscriptions.push(addCppTemplateCommand);

	let createSoitaskForSolutionCommand = vscode.commands.registerCommand(
		"soicode.createSoitaskForSolution",
		createSoitaskFile.bind(undefined, store)
	);
	context.subscriptions.push(createSoitaskForSolutionCommand);

	let createSolutionForSoitaskCommand = vscode.commands.registerCommand(
		"soicode.createSolutionForSoitask",
		createCppFile.bind(undefined, store)
	);
	context.subscriptions.push(createSolutionForSoitaskCommand);

	let taskTreeViewProvider = new TaskTreeViewProvider(store);
	let samplesTreeViewProvider = new SamplesTreeViewProvider(store);

	vscode.window.createTreeView("soicode-tasks", {
		treeDataProvider: taskTreeViewProvider
	});
	vscode.window.createTreeView("soicode-samples", {
		treeDataProvider: samplesTreeViewProvider
	});

	subscribeInsertCppTemplateLens(context, store);
	subscribeInsertSoitaskTemplateLens(context, store);

	let helperBinaryResultReceiver = new HelperBinaryResultReceiver(store);
	helperBinaryResultReceiver.startServer(port);

	store.emitFsChange();

	// await downloadHelperBinary(store);
	await downloadCompilerBundle(store);

	console.log("SOICode has been loaded");
}

async function debugSolutionOnSample() {
	await vscode.window.showErrorMessage(
		"'Debug with sample' is not implemented yet"
	);
}

// async function updateIsSOITaskContext() {
// 	let result = await vscode.workspace.findFiles(".soitask");
// 	console.log(".soitask files: ", result);
// 	let setContextResult = await vscode.commands.executeCommand(
// 		"setContext",
// 		"isSOITask",
// 		result.length >= 1
// 	);
// 	console.log(setContextResult);
// }

// this method is called when your extension is deactivated
export function deactivate() {}

// function sleep(ms: number) {
// 	return new Promise(resolve => {
// 		setTimeout(resolve, ms);
// 	});
// }

// async function createEmptyTask() {
// 	await vscode.window.setStatusBarMessage("Creating new Task");
// 	let solutionName = undefined;
// 	while (true) {
// 		solutionName = await vscode.window.showInputBox({
// 			prompt: "Please select a name for the new Task"
// 		});
// 		if (solutionName !== undefined && /^[a-z0-9A-Z]+$/.test(solutionName)) {
// 			break;
// 		}
// 		let errorMessageClicked = new Promise(async resolve => {
// 			let answer = await vscode.window.showErrorMessage(
// 				"Please choose a name for the new task consisting of alphanumeric symbols",
// 				"Ok",
// 				"Cancel"
// 			);
// 			if (answer === undefined || answer === "Cancel") {
// 				return;
// 			}
// 			resolve();
// 		});
// 		let sleepTimedout = sleep(2000);
// 		await Promise.race([sleepTimedout, errorMessageClicked]);
// 	}
// 	let parentDir;
// 	try {
// 		parentDir = await vscode.window.showOpenDialog({
// 			openLabel: "Create new Task in this Folder",
// 			canSelectFiles: false,
// 			canSelectFolders: true,
// 			canSelectMany: false
// 		});
// 	} catch (err) {
// 		vscode.window.showErrorMessage(err);
// 		return;
// 	}

// 	if (parentDir === undefined || parentDir.length !== 1) {
// 		vscode.window.showErrorMessage("You have to select exactly one directory");
// 		return;
// 	}
// 	let solutionDir = path.join(parentDir[0].fsPath, solutionName);
// 	try {
// 		await fs.mkdir(solutionDir);
// 		const files = await fs.readdir(solutionDir);
// 		if (files.length > 0) {
// 			vscode.window.showErrorMessage("Task directory should be empty");
// 			return;
// 		}
// 	} catch {
// 		vscode.window.showErrorMessage("Directory for new task already exists");
// 	}

// 	let cppFile = path.join(solutionDir, solutionName + ".cpp");
// 	// Write files into new Workdir
// 	try {
// 		// await fs.writeFile(cppFile, cppFileContent(solutionName));
// 		// await fs.mkdir(path.join(solutionDir, ".vscode"));
// 		//await fs.mkdir(path.join(solutionDir, ".vscode", "buildDir"));
// 		// await fs.writeFile(path.join(solutionDir, ".vscode", "launch.json"), launchConfig(solutionName));
// 		// await fs.writeFile(path.join(solutionDir, ".vscode", "tasks.json"), taskJson(solutionName));
// 		// await fs.writeFile(path.join(solutionDir, ".vscode", "c_cpp_properties.json"), cCppProperties(solutionName));
// 		// await fs.writeFile(path.join(solutionDir, ".vscode", "settings.json"), settingsFile(solutionName));
// 		// await fs.writeFile(path.join(solutionDir, ".soicode"), soiCodeFile(solutionName));
// 	} catch (err) {
// 		vscode.window.showErrorMessage(err.toString());
// 	}

// 	// There is a race condition in this code. once opening the folder is done, every extension is killed and gets restarted
// 	vscode.commands.executeCommand(
// 		"vscode.openFolder",
// 		vscode.Uri.parse("file://" + solutionDir)
// 	);
// 	//vscode.workspace.openTextDocument(vscode.Uri.parse('file://' + cppFile));
// 	//vscode.commands.executeCommand('workbench.action.files.openFile', vscode.Uri.parse('file://' + cppFile));
// 	vscode.commands.executeCommand("vscode.open", cppFile);
// 	vscode.commands.executeCommand("workbench.view.explorer");
// }
