import * as vscode from "vscode";
import { selectTaskDefault } from "./soitask";
import {
	getCompilerPath,
	getDebuggerPath,
	getDebuggerFlavor,
	getCompilationFlags,
	getCppHeaderPath,
	getSoiHeaderPath
} from "./compilerBundle";

import { getHelperPath } from "./helperBinary";

import * as fs from "fs-extra";
import { Store } from "./store";
import { taskTreeResourceEntry } from "./taskTreeView";
import {
	getWorkspaceOrShowError,
	getSoitaskFilesOfFulltaskOrError
} from "./storeHelper";
import { selectSample } from "./sample";

export class Runner {
	store: Store;
	port: Number;

	constructor(store: Store, port: Number) {
		this.store = store;
		this.port = port;
	}

	async compile(this: Runner, taskname: string): Promise<"Success" | "Failed"> {
		return new Promise(async resolve => {
			let workspace = getWorkspaceOrShowError(this.store).workspaceDir;
			let soitaskFiles = getSoitaskFilesOfFulltaskOrError(this.store, taskname);

			let compilerPath = getCompilerPath(this.store);

			try {
				await fs.readFile(compilerPath);
			} catch (err) {
				vscode.window.showErrorMessage(
					"Compiler executable not found at expected path"
				);
				this.store.logError(
					"compilation",
					`file at compilerpath '${compilerPath}' not found`,
					err
				);
				resolve("Failed");
			}
			let compileTaskDefinition = {
				label: "build " + taskname + " development",
				// "command": "clang++",
				// "args": ,
				type: "process",
				group: {
					kind: "build",
					isDefault: true
				}
			};

			let headerAsFlag = (path: string) => (path === "" ? [] : ["-I", path]);

			// Todo: allow customizable flags
			let execute = new vscode.ProcessExecution(
				compilerPath,
				[
					soitaskFiles.solutionWorkspacePath!,
					...getCompilationFlags(this.store),
					...headerAsFlag(getCppHeaderPath(this.store)),
					...headerAsFlag(getSoiHeaderPath(this.store)),
					"-o",
					taskname
				],
				{
					cwd: workspace
				}
			);

			try {
				await fs.unlink(soitaskFiles.binaryFsPath);
			} catch (err) {
				if (err.code === "ENOENT") {
					this.store.logInfo(
						"compilation",
						`compiled binary '${compilerPath}' not found`,
						err
					);
				} else {
					this.store.handleError(
						"compilation",
						`deleting binary '${soitaskFiles.binaryFsPath}' failed`,
						err
					);
				}
			}

			// let compileTask = new vscode.Task(compileTaskDefinition, vscode.TaskScope.Workspace, "compile" + taskname, "c++", execute, "$gcc");
			let compileTask = new vscode.Task(
				compileTaskDefinition,
				vscode.TaskScope.Workspace,
				"compile" + taskname,
				"soic",
				execute
			);
			try {
				await vscode.tasks.executeTask(compileTask);
			} catch (err) {
				this.store.logError("compilation", "executing task failed", err);
				resolve("Failed");
			}

			vscode.tasks.onDidEndTask(taskEnded => {
				if (taskEnded.execution.task.name === "compile" + taskname) {
					try {
						fs.readFile(soitaskFiles.binaryFsPath).then(() => {
							resolve("Success");
						});
					} catch (err) {
						this.store.logError(
							"compilation",
							`reading binary '${soitaskFiles.binaryFsPath}" failed`,
							err
						);
						if (err.code !== "EEXIST") {
							resolve("Failed");
						}
					}
				}
			});

			setTimeout(() => {
				this.store.logError("compilation", `timed out compilation`);
				resolve("Failed");
			}, 10000);
		});
	}

	async askAndCompile(
		this: Runner,

		taskname: string | undefined | vscode.Uri | taskTreeResourceEntry
	) {
		let task = await selectTaskDefault(this.store, taskname);

		if ((await this.compile(task)) === "Failed") {
			vscode.window.showErrorMessage("Compilation failed");
			return;
		}
	}

	async runTask(
		this: Runner,
		taskname: string,
		sampleInfo: string,
		sampleargs: string[]
	) {
		if ((await this.compile(taskname)) === "Failed") {
			vscode.window.showErrorMessage("Compilation failed");
			return;
		}

		let workspace = getWorkspaceOrShowError(this.store).workspaceDir;

		let helperPath = getHelperPath(this.store);
		try {
			await fs.readFile(helperPath);
		} catch {
			vscode.window.showErrorMessage(
				"Helper binary not found at expected path"
			);
			return;
		}

		let runTaskDefinition = {
			type: "shell",
			// "command": "clang++",
			// "args": ,
			group: {
				kind: "run",
				isDefault: true
			}
		};
		// Todo: allow customizable flags
		let execute = new vscode.ShellExecution(
			helperPath,
			sampleargs.concat(["--port", "" + this.port]),
			{
				env: {
					SOI_H_EOFCHECK: "1"
				},
				cwd: workspace
			}
		);
		let runTask = new vscode.Task(
			runTaskDefinition,
			vscode.TaskScope.Workspace,
			"run " + taskname + " " + sampleInfo,
			"soicode run",
			execute
		);
		try {
			await vscode.tasks.executeTask(runTask);
		} catch (err) {
			vscode.window.showErrorMessage("Error running sample: " + err.toString());
		}
	}

	async runWithManualInput(
		this: Runner,
		taskname: string | undefined | vscode.Uri | taskTreeResourceEntry
	) {
		taskname = await selectTaskDefault(this.store, taskname);
		let soitaskFile = getSoitaskFilesOfFulltaskOrError(this.store, taskname);
		await this.runTask(taskname, "User Input", [
			"runWithManualInput",
			"--binaryPath",
			soitaskFile.binaryWorkspacePath
		]);
	}
	async runAllSamples(
		this: Runner,
		taskname: string | undefined | vscode.Uri | taskTreeResourceEntry
	) {
		taskname = await selectTaskDefault(this.store, taskname);
		let soitaskFile = getSoitaskFilesOfFulltaskOrError(this.store, taskname);

		await this.runTask(taskname, "All Samples", [
			"runAllSamples",
			"--binaryPath",
			soitaskFile.binaryWorkspacePath,
			"--soitaskPath",
			soitaskFile.soitaskWorkspacePath!
		]);
	}
	async runOneSample(
		this: Runner,
		taskname: string | undefined | vscode.Uri,
		samplename: string | undefined
	) {
		taskname = await selectTaskDefault(this.store, taskname);
		let soitaskFiles = getSoitaskFilesOfFulltaskOrError(this.store, taskname);

		let sample = await selectSample(this.store, taskname, samplename);

		await this.runTask(taskname, "sample", [
			"runOneSample",
			"--binaryPath",
			soitaskFiles.binaryWorkspacePath,
			"--soitaskPath",
			soitaskFiles.soitaskWorkspacePath!,
			"--sampleName",
			sample
		]);
	}

	async debug(
		this: Runner,
		taskname: string,
		extraconfig: any
	): Promise<"Successful" | "Failed"> {
		if ((await this.compile(taskname)) === "Failed") {
			vscode.window.showErrorMessage("Compilation failed");
			return "Failed";
		}

		let workspace = getWorkspaceOrShowError(this.store).workspaceDir;
		let soitaskFiles = getSoitaskFilesOfFulltaskOrError(this.store, taskname);
		let debuggerPath = getDebuggerPath(this.store);
		let debuggerFlavor = getDebuggerFlavor(this.store);

		if (debuggerPath === null) {
			vscode.window.showErrorMessage("Debugger path could not be determined");
			return "Failed";
		}
		try {
			await fs.readFile(debuggerPath);
		} catch {
			vscode.window.showErrorMessage(
				"Debugger executable not found at expected path"
			);
			return "Failed";
		}

		if (debuggerFlavor === null) {
			vscode.window.showErrorMessage("Could not determine type of debugger");
			return "Failed";
		}

		let debugConfig = Object.assign(
			{
				name: "Debug " + taskname,
				type: "cppdbg",
				request: "launch",
				program: soitaskFiles.binaryFsPath,
				MiMode: debuggerFlavor,
				cwd: workspace,
				setupCommands: [
					{
						description: "Enable pretty-printing for debugger",
						text: "-enable-pretty-printing",
						ignoreFailures: true
					}
				]
			},
			extraconfig
		);

		try {
			await fs.readFile(soitaskFiles.binaryFsPath);
		} catch (err) {
			if (err.code !== "EEXIST") {
				vscode.window.showErrorMessage(
					"Unknown error occured when checking binary: " + err.toString()
				);
			} else {
				vscode.window.showErrorMessage(
					"The binary to debug does not exist" + err.toString()
				);
			}
			this.store.handleError(
				"debug",
				`binary path '${soitaskFiles.binaryFsPath}'does not exist`,
				err
			);
		}
		let workspaceFolder = vscode.workspace.getWorkspaceFolder(
			vscode.Uri.file(soitaskFiles.binaryFsPath)
		);

		console.log("Debug parameters:", workspaceFolder, debugConfig);

		let startedSuccessfully = await vscode.debug.startDebugging(
			workspaceFolder,
			debugConfig
		);
		if (startedSuccessfully) {
			return "Successful";
		} else {
			return "Failed";
		}
	}

	async askAndDebugManually(
		this: Runner,
		taskname: string | undefined | vscode.Uri | taskTreeResourceEntry
	) {
		let task = await selectTaskDefault(this.store, taskname);
		if ((await this.debug(task, {})) === "Failed") {
			vscode.window.showErrorMessage("Debugging failed");
			return;
		}
	}
}
