import { parseStoml, StomlError } from "./stoml-parser/parser";

import * as vscode from "vscode";

import * as fs from "fs-extra";
import { Entry } from "./stoml-parser/parser";
import { Store } from "./store";
import { getSoitaskFilesOfFulltaskOrError } from "./storeHelper";
import { fspathToTasknameOrError } from "./soitask";
import { toSortedKeys } from "./util";

// export async function getTaskStoml(taskname: string | null): Promise<Entry[]> {
// 	if (taskname === null) {
// 		let editor = vscode.window.activeTextEditor;
// 		if (editor === undefined) {
// 			throw new Error(
// 				"Can't get stoml if no taskname is passed and no editor is open"
// 			);
// 		} else {
// 			taskname = await filenameToTaskname(editor.document.fileName);
// 		}
// 	}

// 	let soitaskFile = await tasknameToSoitaskFilename(taskname);
// 	let stoml = (await fs.readFile(soitaskFile)).toString();
// 	return parseStoml(stoml);
// }

export function determineSamples(
	soitaskContent: string
): Map<string, Entry> | StomlError {
	try {
		let entries = parseStoml(soitaskContent);
		return new Map(
			entries
				.filter(entry => entry.name[0] === "sample")
				.map(entry => [entry.name.join("."), entry])
		);
	} catch (err) {
		if (err instanceof StomlError) {
			return err;
		}
		throw new Error(
			"Unexpected Error happened when parsing the soitask file: " +
				err.toString()
		);
	}
}

export async function listSamplesForTaskname(
	store: Store,
	taskname: string
): Promise<Map<string, Entry>> {
	let path: string;
	try {
		path = getSoitaskFilesOfFulltaskOrError(
			store,
			taskname
		).soitaskFsPath!.toString();
	} catch (err) {
		return new Map();
	}
	let content: string;
	try {
		content = (await fs.readFile(path)).toString();
	} catch (err) {
		vscode.window.showErrorMessage("Cannot read soitask file");
		throw new Error("Cannot read soitask file");
	}

	let sampleList = determineSamples(content);
	if (sampleList instanceof StomlError) {
		vscode.window.showErrorMessage(
			"Cannot parse soitask. It has the following error: " +
				sampleList.toString()
		);
		throw new Error("Cannot read soitask file");
	}
	return sampleList;
}

export async function selectSample(
	store: Store,
	taskname: string,
	sample: string | undefined | null
): Promise<string> {
	if (typeof sample === "string") {
		return sample;
	}
	let sampleList = await listSamplesForTaskname(store, taskname);
	sample = await vscode.window.showQuickPick(toSortedKeys(sampleList), {
		canPickMany: false,
		placeHolder: "sample.mysample"
	});
	if (sample === undefined) {
		vscode.window.showErrorMessage("No sample has been selected");
		throw new Error("No sample has been selected");
	}
	return sample;
}

export async function gotoSample(
	store: Store,
	sampleName: string | null | undefined
) {
	let editor = vscode.window.activeTextEditor;
	if (editor === undefined) {
		vscode.window.showErrorMessage("No file open");
		throw new Error("No file open");
	}
	let taskname = fspathToTasknameOrError(store, editor.document.fileName);
	sampleName = await selectSample(store, taskname, sampleName);
	let sample = store.samples.get(sampleName);
	if (sample === undefined) {
		vscode.window.showErrorMessage("The selected sample does not exist");
		throw new Error("The selected sample does not exist");
	}

	let loc = new vscode.Range(
		new vscode.Position(sample.span.startLine, 0),
		new vscode.Position(
			sample.span.startLine,
			editor.document.lineAt(sample.span.startLine).range.end.character
		)
	);

	let openPath = getSoitaskFilesOfFulltaskOrError(store, taskname)
		.soitaskFsPath!;
	let doc = await vscode.workspace.openTextDocument(openPath);
	editor = await vscode.window.showTextDocument(doc);
	editor.revealRange(loc);
	editor.selection = new vscode.Selection(
		new vscode.Position(sample.span.startLine, 0),
		new vscode.Position(
			sample.span.startLine,
			editor.document.lineAt(sample.span.startLine).range.end.character
		)
	);
}
