import * as vscode from "vscode";

import * as path from "path";
import * as fs from "fs-extra";

import { taskTreeResourceEntry } from "./taskTreeView";
import { Store } from "./store";
import { getWorkspaceOrShowError, listFullTasksOrError } from "./storeHelper";
import { executableFileEnding } from "./util";

export type SoitaskFiles = {
	taskname: string; // Relative to workspace, no file ending
	solutionFsPath: string | null;
	soitaskFsPath: string | null;
	binaryFsPath: string; // No guarantee the file actually exists, but if it exists, this is the path
	solutionWorkspacePath: string | null;
	soitaskWorkspacePath: string | null;
	binaryWorkspacePath: string; // No guarantee the file actually exists, but if it exists, this is the path
};
export function fspathToTasknameStringOnly(
	workspace: string,
	fspath: string
): string {
	return path.relative(
		workspace,
		path.join(path.dirname(fspath), path.basename(fspath, path.extname(fspath)))
	);
}
export async function listSoitasks(
	workspace: string
): Promise<Map<string, SoitaskFiles>> {
	let tasks: Map<string, SoitaskFiles> = new Map();
	let cppFiles = await vscode.workspace.findFiles("**/*.cpp");
	let soitaskFiles = await vscode.workspace.findFiles("**/*.soitask");
	let tasknames = [];
	let cppTasknames = new Map<string, string>();
	let soitaskTasknames = new Map<string, string>();

	for (let cppFile of cppFiles) {
		let taskname = fspathToTasknameStringOnly(workspace, cppFile.fsPath);
		tasknames.push(taskname);
		cppTasknames.set(taskname, cppFile.fsPath);
	}
	for (let soitaskFile of soitaskFiles) {
		let taskname = fspathToTasknameStringOnly(workspace, soitaskFile.fsPath);
		tasknames.push(taskname);
		soitaskTasknames.set(taskname, soitaskFile.fsPath);
	}
	tasknames = Array.from(new Set(tasknames));

	const undefinedToNullMapString = function(
		a: undefined | string
	): string | null {
		return a === undefined ? null : a;
	};
	const toRelativePath = function(a: undefined | string): string | null {
		return a === undefined ? null : path.join(path.relative(workspace, a));
	};

	for (let taskname of tasknames) {
		tasks.set(taskname, {
			taskname: taskname,
			solutionFsPath: undefinedToNullMapString(cppTasknames.get(taskname)),
			soitaskFsPath: undefinedToNullMapString(soitaskTasknames.get(taskname)),
			binaryFsPath: path.join(workspace, taskname + executableFileEnding()),
			solutionWorkspacePath: toRelativePath(cppTasknames.get(taskname)),
			soitaskWorkspacePath: toRelativePath(soitaskTasknames.get(taskname)),
			binaryWorkspacePath: taskname + executableFileEnding()
		});
	}
	return tasks;
}

export function fspathToTasknameOrError(store: Store, fspath: string): string {
	if (
		path.extname(fspath) !== ".cpp" &&
		path.extname(fspath) !== ".soitask" &&
		path.extname(fspath) !== ""
	) {
		throw new Error("Extension needs to be .cpp or .soitask");
	}
	return path.relative(
		getWorkspaceOrShowError(store).workspaceDir,
		path.join(path.dirname(fspath), path.basename(fspath, path.extname(fspath)))
	);
}

export async function createCppFile(store: Store, taskname: string) {
	taskname = path.basename(taskname);
	let preExistingTask = getWorkspaceOrShowError(store).task.get(taskname);
	if (
		preExistingTask !== undefined &&
		preExistingTask.solutionFsPath !== null
	) {
		vscode.window.showErrorMessage("C++ file already exists");
		throw new Error("C++ file already exists");
	}
	let workspace = getWorkspaceOrShowError(store).workspaceDir;
	try {
		await fs.writeFile(path.join(workspace, taskname + ".cpp"), "");

		store.emitFsChange();
	} catch (err) {
		vscode.window.showErrorMessage("Cannot create .cpp file");
		throw new Error("Cannot create .cpp file");
	}
}

export async function createSoitaskFile(store: Store, taskname: string) {
	taskname = path.basename(taskname);
	let preExistingTask = getWorkspaceOrShowError(store).task.get(taskname);
	if (preExistingTask !== undefined && preExistingTask.soitaskFsPath !== null) {
		vscode.window.showErrorMessage("Soitask file already exists");
		throw new Error("Soitask file already exists");
	}
	let workspace = getWorkspaceOrShowError(store).workspaceDir;
	try {
		await fs.writeFile(path.join(workspace, taskname + ".soitask"), "");
		store.emitFsChange();
	} catch (err) {
		vscode.window.showErrorMessage("Cannot create soitask file");
		throw new Error("Cannot create soitask file");
	}
}
export async function createTask(store: Store) {
	let taskname = await vscode.window.showInputBox({
		placeHolder: "mytask",
		prompt: "Enter the name of the new task"
	});
	if (taskname === undefined) {
		return;
	}
	if (taskname.length === 0) {
		vscode.window.showErrorMessage("Task name cannot be empty");
		await createTask(store);
		return;
	}
	await createCppFile(store, taskname);
	await createSoitaskFile(store, taskname);
}

async function selectTask(store: Store): Promise<string> {
	let task = await vscode.window.showQuickPick(
		listFullTasksOrError(store).map(soitaskFiles => soitaskFiles.taskname),
		{
			canPickMany: false,
			placeHolder: "mytask"
		}
	);
	if (task === undefined) {
		vscode.window.showErrorMessage("No task has been selected");
		throw new Error("No task has been selected");
	}
	return task;
}

export async function selectTaskDefault(
	store: Store,
	defaultTask: string | undefined | null | vscode.Uri | taskTreeResourceEntry
): Promise<string> {
	let editor = vscode.window.activeTextEditor;
	if (defaultTask === undefined || defaultTask === null) {
		if (editor !== undefined) {
			try {
				return fspathToTasknameOrError(store, editor.document.uri.fsPath);
			} catch (err) {}
		}
		return await selectTask(store);
	} else if (defaultTask instanceof vscode.Uri) {
		return fspathToTasknameOrError(store, defaultTask.fsPath);
	} else if (typeof defaultTask === "string") {
		return defaultTask;
	} else {
		return defaultTask.taskname;
	}
}
