import * as vscode from "vscode";

import * as path from "path";
import * as fs from "fs-extra";
import { SoitaskFiles, listSoitasks } from "./soitask";
import { Store } from "./store";
import {
	WorkspaceSettings,
	soiWorkspaceTemplate_0_8_0,
	checkIsWorkspaceSettings
} from "./soiworkspaceFormats/soiworkspace-0-8-0";

export type SoiWorkspace = {
	task: Map<string, SoitaskFiles>;
	settings: WorkspaceSettings;
	workspaceDir: string;
};

export type WorkspaceState = "NoFolderOpen" | "NoSOIWorkspace" | SoiWorkspace;

export const emptyWorkspaceState: WorkspaceState = "NoFolderOpen";

async function determineWorkspaceFolder(): Promise<string | null> {
	let dirs = vscode.workspace.workspaceFolders;
	if (dirs === undefined || dirs.length <= 0 || dirs.length > 1) {
		return null;
	}
	return dirs[0].uri.fsPath;
}

async function determineWorkspaceSettingsFileContent(
	store: Store,
	workspace: string
): Promise<string | null> {
	let workspaceSettings = await vscode.workspace.findFiles("soiworkspace.json");
	if (workspaceSettings.length >= 2 || workspaceSettings.length < 0) {
		// Todo: Replace with assert
		throw new Error("This is impossible");
	}
	if (workspaceSettings.length === 0) {
		return null;
	}
	try {
		return (
			await fs.readFile(path.join(workspace, "soiworkspace.json"))
		).toString();
	} catch (err) {
		vscode.window.showErrorMessage(
			"Cannot read soi workspace settings file: " + err.toString()
		);
		return null;
	}
}

async function createWorkspaceSettingsFileContent(
	store: Store,
	workspace: string
): Promise<"failed" | "success"> {
	let workspaceSettings = await vscode.workspace.findFiles("soiworkspace.json");
	if (workspaceSettings.length >= 2 || workspaceSettings.length < 0) {
		// Todo: Replace with assert
		// throw new Error("This is impossible");
		return "failed";
	}
	if (workspaceSettings.length === 0) {
		try {
			await fs.writeFile(
				path.join(workspace, "soiworkspace.json"),
				soiWorkspaceTemplate_0_8_0(store)
			);
		} catch (err) {
			vscode.window.showErrorMessage(
				"Error initializing workspace: " + err.toString()
			);
			return "failed";
		}
	}
	if ((await determineWorkspaceSettings(store, workspace)) === null) {
		return "failed";
	}
	return "success";
}

async function determineWorkspaceSettings(
	store: Store,
	workspace: string
): Promise<WorkspaceSettings | null> {
	let workspaceSettingsFileContent = await determineWorkspaceSettingsFileContent(
		store,
		workspace
	);
	if (workspaceSettingsFileContent === null) {
		return null;
	}
	try {
		return checkIsWorkspaceSettings(
			JSON.parse(workspaceSettingsFileContent.toString())
		);
	} catch (e) {
		vscode.window.showErrorMessage(
			"Cannot parse workspace settings file: " + e.toString()
		);
		return null;
	}
}

export async function readWorkspaceState(
	store: Store
): Promise<WorkspaceState> {
	let workdir = await determineWorkspaceFolder();
	if (workdir === null) {
		return "NoFolderOpen";
	}
	let settings = await determineWorkspaceSettings(store, workdir);
	if (settings === null) {
		return "NoSOIWorkspace";
	}
	let tasks = await listSoitasks(workdir);

	return {
		settings: settings,
		task: tasks,
		workspaceDir: workdir
	};
}

export async function initializeWorkspace(store: Store) {
	let workspace = await determineWorkspaceFolder();
	if (workspace !== null) {
		let result = await createWorkspaceSettingsFileContent(store, workspace);
		if (result === "failed") {
			vscode.window.showErrorMessage("Couldn't initialize new workspace");
		}
		store.emitFsChange();
	}
}
