import * as fs from "fs-extra";

interface Span {
	startLine: number;
	startPos: number;
	endLine: number;
	endPos: number;
}

export class StomlError extends Error {
	constructor(public span: Span, message: string) {
		super(message);
		Object.setPrototypeOf(this, new.target.prototype); // restore prototype chain
	}
}

interface KeyValue {
	key: string;
	value: string;
	span: Span;
}

export interface Entry {
	name: [string, string];
	span: Span;
	data: KeyValue[];
}

enum TokenKind {
	WS,
	LBrac,
	RBrac,
	Iden,
	Dot,
	Eq,
	Str,
	RawStr,
	EOF
}

class Token {
	constructor(public kind: TokenKind, public span: Span, public raw: string) {}
}

class Tokenizer {
	private index: number;
	private lineNr: number;
	private linePos: number;
	private current: Token | null;

	constructor(private code: string) {
		this.index = 0;
		this.lineNr = 0;
		this.linePos = 0;
		this.current = null;
	}
	public position(): [number, number] {
		return [this.lineNr, this.linePos];
	}
	public spanFrom(start: [number, number]): Span {
		return {
			startLine: start[0],
			startPos: start[1],
			endLine: this.lineNr,
			endPos: this.linePos
		};
	}
	private peekChar(): string | null {
		if (this.index === this.code.length) {
			return null;
		}
		return this.code[this.index];
	}
	private nextChar(): string | null {
		if (this.index === this.code.length) {
			return null;
		}
		let c = this.code[this.index];
		this.index += 1;
		if (c === "\n") {
			this.lineNr += 1;
			this.linePos = 0;
		} else {
			this.linePos += 1;
		}
		return c;
	}
	public peek(): Token {
		if (this.current === null) {
			this.current = this.next();
		}
		return this.current;
	}
	private consumeUntil(pred: (x: string) => boolean) {
		while (true) {
			let x = this.peekChar();
			if (x === null || pred(x)) {
				return;
			}
			this.nextChar();
		}
	}
	private consumeNext(): TokenKind {
		let startLine = this.lineNr;
		let startPos = this.linePos;
		let c = this.nextChar();
		if (c === null) {
			return TokenKind.EOF;
		}
		if (/\s/.test(c)) {
			return TokenKind.WS;
		}
		if (c === "#") {
			this.consumeUntil(c => c === "\n");
			this.nextChar();
			return TokenKind.WS;
		}
		if (c === "[") {
			return TokenKind.LBrac;
		}
		if (c === "]") {
			return TokenKind.RBrac;
		}
		if (c === ".") {
			return TokenKind.Dot;
		}
		if (c === "=") {
			return TokenKind.Eq;
		}
		if (/\w/.test(c)) {
			this.consumeUntil(c => !/\w/.test(c));
			return TokenKind.Iden;
		}
		if (c === "'" || c === '"') {
			let delim = c;
			c = this.nextChar();
			if (c !== delim) {
				let isEsc = false;
				this.consumeUntil(c => {
					if (isEsc) {
						isEsc = false;
						return false;
					}
					if (c === "\\") {
						isEsc = true;
					} else {
						isEsc = false;
						if (c === delim) {
							return true;
						}
					}
					return false;
				});
				if (this.nextChar() === null) {
					throw new StomlError(
						{
							startLine: startLine,
							startPos: startPos,
							endLine: this.lineNr,
							endPos: this.linePos
						},
						"Unexpected EOF in string literal"
					);
				}
				return TokenKind.Str;
			} else {
				if (this.peekChar() !== delim) {
					return TokenKind.Str; // empty string
				}
				this.nextChar();
				let cnt = 0;
				this.consumeUntil(c => {
					if (c === delim) {
						cnt += 1;
					} else {
						cnt = 0;
					}
					return cnt === 3;
				});
				if (this.nextChar() === null) {
					throw new StomlError(
						{
							startLine: startLine,
							startPos: startPos,
							endLine: this.lineNr,
							endPos: this.linePos
						},
						"Unexpected EOF in string literal"
					);
				}
				return TokenKind.RawStr;
			}
		}
		throw new StomlError(
			{
				startLine: startLine,
				startPos: startPos,
				endLine: this.lineNr,
				endPos: this.linePos
			},
			"Invalid Character: `" + c + "'"
		);
	}
	public next(): Token {
		if (this.current !== null) {
			let ans = this.current;
			this.current = null;
			return ans;
		}
		let startLine = this.lineNr;
		let startPos = this.linePos;
		let startIndex = this.index;
		let kind = this.consumeNext();
		let endLine = this.lineNr;
		let endPos = this.linePos;
		let endIndex = this.index;
		return new Token(
			kind,
			{
				startLine: startLine,
				startPos: startPos,
				endLine: endLine,
				endPos: endPos
			},
			this.code.substr(startIndex, endIndex - startIndex)
		);
	}
}

function stomlSkipWS(t: Tokenizer) {
	while (true) {
		let tok = t.peek();
		if (tok.kind !== TokenKind.WS) {
			break;
		}
		t.next();
	}
}

function stomlExpect(t: Tokenizer, kind: TokenKind): Token {
	let tok = t.next();
	if (tok.kind !== kind) {
		throw new StomlError(tok.span, "Expected '" + TokenKind[kind] + "'");
	}
	return tok;
}

function stomlExpectAllowWS(t: Tokenizer, kind: TokenKind): Token {
	stomlSkipWS(t);
	return stomlExpect(t, kind);
}

function stomlUnescape(str: string, delim: string): string {
	let unescaped: string[] = [];
	let isEsc = false;
	for (var i = 0; i < str.length; i++) {
		let c = str[i];
		if (isEsc) {
			if (c === "n") {
				unescaped.push("\n");
			}
			unescaped.push(c);
			isEsc = false;
		} else if (c === "\\") {
			isEsc = true;
		} else {
			unescaped.push(c);
			isEsc = false;
		}
	}
	return unescaped.join("");
}

function stomlParseKeyVal(t: Tokenizer): KeyValue | null {
	stomlSkipWS(t);
	let start = t.position();
	if (t.peek().kind !== TokenKind.Iden) {
		return null;
	}
	let tkKey = t.next();
	stomlExpectAllowWS(t, TokenKind.Eq);
	stomlSkipWS(t);
	let tkVal = t.next();
	let val;
	if (tkVal.kind === TokenKind.Str) {
		val = stomlUnescape(
			tkVal.raw.substr(1, tkVal.raw.length - 2),
			tkVal.raw[0]
		);
	} else if (tkVal.kind === TokenKind.RawStr) {
		// toml spec: A newline immediately following the opening delimiter will be trimmed.
		let offset = tkVal.raw[3] === "\n" ? 1 : 0;
		val = tkVal.raw.substr(3 + offset, tkVal.raw.length - 6 - offset);
	} else {
		throw new StomlError(tkVal.span, "Expected string");
	}
	let span = t.spanFrom(start);
	return {
		key: tkKey.raw,
		value: val,
		span: span
	};
}

function stomlParseEntry(t: Tokenizer): Entry | null {
	stomlSkipWS(t);
	let start = t.position();
	if (t.peek().kind === TokenKind.EOF) {
		return null;
	}
	stomlExpectAllowWS(t, TokenKind.LBrac);
	let n1 = stomlExpectAllowWS(t, TokenKind.Iden);
	stomlExpectAllowWS(t, TokenKind.Dot);
	let n2 = stomlExpectAllowWS(t, TokenKind.Iden);
	stomlExpectAllowWS(t, TokenKind.RBrac);
	let data: KeyValue[] = [];
	while (true) {
		let kv = stomlParseKeyVal(t);
		if (kv === null) {
			break;
		}
		data.push(kv);
	}
	let span = t.spanFrom(start);
	return { name: [n1.raw, n2.raw], span: span, data: data };
}

export function parseStoml(s: string): Entry[] {
	let t = new Tokenizer(s);
	var entries: Entry[] = [];
	while (true) {
		let entry = stomlParseEntry(t);
		if (entry === null) {
			break;
		}
		entries.push(entry);
	}
	return entries;
}

function run(s: string) {
	try {
		console.log(JSON.stringify(parseStoml(s)));
	} catch (e) {
		console.log(e);
	}
}

export function test() {
	fs.readdir("./", (err, files) => {
		files.forEach(file => {
			if (file.endsWith(".stoml")) {
				console.log(file);
				var s: string = fs.readFileSync(file, "utf8");
				run(s);
			}
		});
	});
}
