import * as vscode from "vscode";
import { Entry } from "./stoml-parser/parser";
import {
	WorkspaceState,
	emptyWorkspaceState,
	readWorkspaceState
} from "./soiworkspace";
import { SampleExecutionResult } from "./helperBinaryResultReceiver";
import { listSamplesForTaskname } from "./sample";
import { fspathToTasknameStringOnly } from "./soitask";

type SampleEvaluationLabel = "Unevaluated" | "Correct" | "Grading" | "Wrong";

type SampleEvaluation = {
	exec: SampleExecutionResult[];
	correctness: SampleEvaluationLabel;
};

export class Store {
	evaluations: Map<[string, string], SampleEvaluation>;
	samples: Map<string, Entry>;
	workspace: WorkspaceState;
	soicodeExtensionDataPath: string;
	soicodeExtensionImagePath: string;
	soicodeExtensionDistPath: string;
	fileChangeSubscribers: (() => void)[];
	allSampleChangeSubscribers: (() => void)[];
	sampleUpdateSubscriber: ((s: string) => void)[];
	private channel: vscode.OutputChannel;

	constructor(context: vscode.ExtensionContext) {
		this.channel = vscode.window.createOutputChannel("soicode");
		context.subscriptions.push(this.channel);
		this.soicodeExtensionDataPath = context.globalStoragePath;
		this.soicodeExtensionImagePath = context.asAbsolutePath("images");
		this.soicodeExtensionDistPath = context.asAbsolutePath("dist");
		this.evaluations = new Map();
		this.workspace = emptyWorkspaceState;
		this.samples = new Map();
		this.fileChangeSubscribers = [];
		this.allSampleChangeSubscribers = [];
		this.sampleUpdateSubscriber = [];
		this.subscribeStoreToEvents(context);
	}

	handleError(this: Store, location: string, info: string, err: Error): never {
		this.channel.appendLine(`#### Error ${location} ####`);
		this.channel.appendLine(info);
		this.channel.appendLine("error name:");
		this.channel.appendLine(err.name);
		this.channel.appendLine("message:");
		this.channel.appendLine(err.message);
		if (err.stack !== undefined) {
			this.channel.appendLine("stack trace");
			this.channel.appendLine(err.stack);
		}
		this.channel.show();
		throw err;
	}
	logInfo(this: Store, location: string, info: string, err?: Error) {
		this.channel.appendLine(`#### info at [${location}] ####`);
		this.channel.appendLine(info);
		if (err !== undefined) {
			this.channel.appendLine(err.name);
			this.channel.appendLine("message:");
			this.channel.appendLine(err.message);
			if (err.stack !== undefined) {
				this.channel.appendLine("stack trace");
				this.channel.appendLine(err.stack);
			}
		}
	}
	logError(this: Store, location: string, info: string, err?: Error) {
		this.channel.appendLine(`#### error at [${location}] ####`);
		this.channel.appendLine(info);
		if (err !== undefined) {
			this.channel.appendLine(err.name);
			this.channel.appendLine("message:");
			this.channel.appendLine(err.message);
			if (err.stack !== undefined) {
				this.channel.appendLine("stack trace");
				this.channel.appendLine(err.stack);
			}
		}
	}

	addExecutionResult(
		this: Store,
		taskname: string,
		samplename: string,
		res: SampleExecutionResult
	) {
		let previous = this.evaluations.get([taskname, samplename]);
		if (previous === undefined) {
			throw new Error("this sample doesn't exist");
		}
		previous.exec.push(res);
		this.evaluations.set([taskname, samplename], previous);
	}

	emitFsChange(this: Store) {
		readWorkspaceState(this).then(workspaceState => {
			this.workspace = workspaceState;
			for (let sub of this.fileChangeSubscribers) {
				sub();
			}
			this.emitAllSamplesChange();
		});
	}
	emitAllSamplesChange(this: Store) {
		let editor = vscode.window.activeTextEditor;
		let workspace = this.workspace;

		let updateIt = (store: Store) => {
			for (let sub of this.allSampleChangeSubscribers) {
				sub();
			}
		};

		if (
			editor === undefined ||
			workspace === "NoFolderOpen" ||
			workspace === "NoSOIWorkspace"
		) {
			this.samples = new Map();
			updateIt(this);
			return;
		}
		let taskname = fspathToTasknameStringOnly(
			workspace.workspaceDir,
			editor.document.fileName
		);
		if (workspace.task.get(taskname) === undefined) {
			this.samples = new Map();
			updateIt(this);
			return;
		}
		listSamplesForTaskname(this, taskname).then(samples => {
			this.samples = samples;

			updateIt(this);
		});
	}
	emitSampleUpdate(this: Store, sample: string) {
		for (let sub of this.sampleUpdateSubscriber) {
			sub(sample);
		}
	}

	subscribeStoreToEvents(this: Store, context: vscode.ExtensionContext) {
		let fswatcher = vscode.workspace.createFileSystemWatcher(
			"**/*",
			false,
			false,
			false
		);
		fswatcher.onDidChange(e => {
			this.emitFsChange();
		});
		fswatcher.onDidCreate(e => {
			this.emitFsChange();
		});
		fswatcher.onDidDelete(e => {
			this.emitFsChange();
		});

		context.subscriptions.push(
			vscode.workspace.onDidSaveTextDocument(doc => {
				this.emitFsChange();
			}),
			vscode.workspace.onDidOpenTextDocument(doc => {
				this.emitAllSamplesChange();
			}),
			vscode.window.onDidChangeActiveTextEditor(doc => {
				this.emitAllSamplesChange();
			})
		);
		setInterval(function(this: Store) {
			this.emitFsChange();
		}, 30000);
	}
}
