import * as vscode from "vscode";
import * as path from "path";
import { Store } from "./store";
import {
	listFullTasksOrError,
	getSoitaskFilesOfFulltaskOrError
} from "./storeHelper";

type Task = {
	type: "task";
	taskname: string;
	title: string;
	children: Entry[];
};
type Soitask = { type: "soitask"; taskname: string; title: string };
type Solution = { type: "solution"; taskname: string; title: string };
type Command = {
	type: "command";
	title: string;
	vsCodeCommand: vscode.Command;
};
export type Entry = Task | Soitask | Solution | Command;
export type taskTreeResourceEntry = Task | Soitask | Solution;

function generateEntries(store: Store): Entry[] {
	if (store.workspace === "NoFolderOpen") {
		return [
			{
				type: "command",
				title: "Vscode open folder",
				vsCodeCommand: {
					title: "Vscode open folder",
					command: "workbench.action.files.openLocalFolder",
					tooltip: "Currently in Vscode no folder is open"
				}
			}
		];
	}
	if (store.workspace === "NoSOIWorkspace") {
		return [
			{
				type: "command",
				title: "Initialize Workspace",
				vsCodeCommand: {
					title: "Initialize Workspace",
					command: "soicode.initializeWorkspace",
					tooltip:
						"Currently the workspace has not been initialized to be used with SOI Code. Let's initialize it"
				}
			}
		];
	}
	let tasks = listFullTasksOrError(store);
	if (tasks.length === 0) {
		return [
			{
				type: "command",
				title: "Create new Task",
				vsCodeCommand: {
					title: "Create new task",
					command: "soicode.createTask",
					tooltip:
						"Currently no task exists in the workspace. Let's create a new one"
				}
			}
		];
	}
	let addSolution = (taskname: string): Command => {
		return {
			type: "command",
			title: "Create a cpp solution for this task",
			vsCodeCommand: {
				title: "Create a cpp solution for this task",
				command: "soicode.createSolutionForSoitask",
				arguments: [taskname],
				tooltip:
					"Currently the task '" +
					taskname +
					"' doesn't have a C++ solution file. Let's add it."
			}
		};
	};
	const addSoitask = (taskname: string): Command => {
		return {
			type: "command",
			title: "Create a Soicode file for this task",
			vsCodeCommand: {
				title: "Create a Soicode file for this task",
				command: "soicode.createSoitaskForSolution",
				arguments: [taskname],
				tooltip:
					"Currently the task '" +
					taskname +
					"' doesn't have a soitask file. Let's add it."
			}
		};
	};

	return tasks.map(
		(elem): Task => {
			return {
				type: "task",
				title: elem.taskname,
				taskname: elem.taskname,
				children: [
					elem.solutionWorkspacePath === null
						? addSolution(elem.taskname)
						: {
								type: "solution",
								title: elem.solutionWorkspacePath,
								taskname: elem.taskname
						  },
					elem.soitaskWorkspacePath === null
						? addSoitask(elem.taskname)
						: {
								type: "soitask",
								title: elem.soitaskWorkspacePath,
								taskname: elem.taskname
						  }
				]
			};
		}
	);
}

export class TaskTreeViewProvider implements vscode.TreeDataProvider<Entry> {
	store: Store;
	private entries: Entry[];

	triggerUpdate = new vscode.EventEmitter<Entry | undefined>();
	onDidChangeTreeData = this.triggerUpdate.event;

	constructor(store: Store) {
		this.store = store;
		this.entries = generateEntries(store);
		this.store.fileChangeSubscribers.push(
			this.updateData.bind(this, undefined)
		);
	}

	async getTreeItem(
		this: TaskTreeViewProvider,
		elem: Entry
	): Promise<vscode.TreeItem> {
		let item = new vscode.TreeItem(
			elem.title,
			elem.type === "task"
				? vscode.TreeItemCollapsibleState.Collapsed
				: vscode.TreeItemCollapsibleState.None
		);
		item.contextValue = elem.type;
		// item.contextValue = "test";

		switch (elem.type) {
			case "command":
				item.iconPath = path.join(
					this.store.soicodeExtensionImagePath,
					"soicode-activitybar-dark.svg"
				);
				item.command = elem.vsCodeCommand;
				break;
			case "task":
				item.iconPath = {
					light: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/light/task.svg"
					),
					dark: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/dark/task.svg"
					)
				};

				break;
			case "solution":
				item.iconPath = {
					light: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/light/filetype-cpp.svg"
					),
					dark: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/dark/filetype-cpp.svg"
					)
				};
				item.command = {
					command: "vscode.open",
					title: "Edit this file", //This is ignored
					tooltip: "Edit this file", //This is ignored
					arguments: [
						vscode.Uri.file(
							getSoitaskFilesOfFulltaskOrError(this.store, elem.taskname)
								.solutionFsPath!
						)
					]
				};
				break;
			case "soitask":
				item.iconPath = {
					light: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/light/sample.svg"
					),
					dark: path.join(
						this.store.soicodeExtensionImagePath,
						"./icons/dark/sample.svg"
					)
				};
				item.command = {
					command: "vscode.open",
					title: "Edit this file",
					tooltip: "Edit this file",
					arguments: [
						vscode.Uri.file(
							getSoitaskFilesOfFulltaskOrError(this.store, elem.taskname)
								.soitaskFsPath!
						)
					]
				};
				break;
		}
		// 	// Todo: delete this?
		// if (elem.type !== "command") {
		// 	item.resourceUri = vscode.Uri.file(
		// 		path.join(workspace!, elem.relativeName)
		// 	);
		// }
		return item;
	}

	async updateData(this: TaskTreeViewProvider) {
		this.entries = generateEntries(this.store);
		this.triggerUpdate.fire(undefined);
	}

	getChildren(this: TaskTreeViewProvider, elem?: Entry): Entry[] {
		if (elem === undefined) {
			return this.entries;
		} else {
			if (
				elem.type === "solution" ||
				elem.type === "soitask" ||
				elem.type === "command"
			) {
				return [];
			}
			return elem.children;
		}
	}
}
